# perl interface for jags (just another Gibbs sampler)
# '17/07/05 by KGK

package Makeshift::Jags;

use Makeshift::Jags::var;
use Makeshift::Jags::vars;
use Makeshift::Jags::coda;

use Carp::Assert;
use Mouse;
use IO::File;
use File::Temp qw/ tempfile tempdir /;

has 'jags_command' => (is => 'rw', default => "jags");
has 'tmproot' => (is => 'ro', default => "/tmp");
has 'CLEANUP' => (is => 'ro', default => 1);
has 'tmpdir' => (is => 'ro', lazy => 1,
				 default => sub {my $self = shift;
								 return tempdir(DIR => $self->tmproot,
												CLEANUP => $self->CLEANUP)});
has ['jags_file', 'data_file', 'init_file']
  => (is => 'ro', lazy => 1, builder => '_handle_name');
has ['model_file', 'param_file', 'sampler_file'] => (is => 'rw');
has ['data', 'param'] => (is => 'ro',
						  default => sub{Makeshift::Jags::vars->new});
has '_monitor' => (is => 'ro', lazy_build => 1);
has 'nchains' => (is => 'rw', default => 1);
has 'adapt' => (is => 'rw', default => 0);
has ['burnin', 'update'] => (is => 'rw', default => 1000);
has 'coda' => (is => 'ro', lazy_build => 1);
has 'coda_arg' => (is => 'rw', default => "*");
has 'coda_dir' => (is => 'rw', lazy => 1,
				   default => sub {return shift->tmpdir});
has 'script' => (is => 'ro', lazy_build => 1);

__PACKAGE__->meta->make_immutable();

# read_csv(file): read data from CSV file
# file: comment line: start by '#'
#       first line: list of variable names
sub read_csv {
	my $self = shift;
	my $file = shift;
	my @var = ();
	my %vals = ();

	assert($file, "file to read must be specified");
	assert(-f $file, "$file is not a file");
	assert(-r $file, "$file is not readable");

	open(CSV, "<", $file) or die "$!";
	while (<CSV>) {
		next if /^#/;
		chomp;
		s/^\s+//;
		if (@var == 0) {
			s/"//g;
			@var = split(/,\s*/);
			for (@var) {
				$vals{$_} = [];
			}
		} else {
			my $i = 0;
			for my $val (split(/,\s*/)) {
				push @{$vals{$var[$i++]}}, $val;
			}
		}
	}
	close(CSV);

	for (@var) {
		$self->data->set($_ => $vals{$_});
	}

	return scalar @{$vals{$var[0]}};
}

# set_monitor(var, opt): set monitoring parameter
# var: name of parameter
# opt[optional]: option of monitor
sub set_monitor {
    my $self = shift;
	my $var = shift;
	my $opt = shift;

	$self->_monitor->{$var} = $opt;

    return $self;
}

# del_monitor(var1, var2,...): delete from monitoring parameter list
# var: name of parameter
sub del_monitor {
    my $self = shift;

	assert(@_ > 0, "monitoring parameter must be specified");

	for (@_) {
		delete $self->_monitor->{$_};
	}

    return $self;
}

# save_script(modelfile, jagsfile): write jags script file
# modelfile[optional]: model file
# jagsfile[optional]: jags script file
sub save_script {
    my $self = shift;
	my $modelfile = shift;
	my $jagsfile = shift;

	$self->model_file($modelfile) if $modelfile;
	assert($self->model_file, "file name for model must be specified");
	assert($self->jags_file, "file name for jags-script must be specified");

	$self->data->save($self->data_file->{handle})
	  unless -s $self->data_file->{name};
	$self->param->save($self->init_file->{handle})
	  unless -s $self->init_file->{name};

	my $fh = $self->jags_file->{handle};
	for (@{$self->script}) {
		print $fh $_;
	}

    return $self->jags_file->{name};
}

# run(num): run jags
# num[optional]: number of update iterations
sub run {
    my $self = shift;
	my $num = shift;

	$self->update($num) if $num;
	system $self->jags_command . " " . $self->save_script;

	die $self->jags_command . " return $?" if $?;

    return $self;
}

# _handle_name: return hash reference for handle and name for tmpfile
sub _handle_name {
	my $self = shift;
	my ($fh, $fn) = tempfile(DIR => $self->tmpdir);

	$fh->autoflush(1);

	return {handle => $fh, name => $fn};
}

# _build__monitor: default initialize of '_monitor'
# all parameters are monitored as default
sub _build__monitor {
	my $self = shift;
	my %mon = ();

	for ($self->param->get_array) {
		$mon{$_->name} = undef;
	}

	return \%mon;
}

# _build_script: default initialize of 'script'
# create standard jags script as strings array
sub _build_script {
    my $self = shift;
	my @sc = ();
	my $burnin = ($self->adapt) ? "adapt" : "update";
	my $stem = ($self->coda_dir)
	  ? ", stem(\"" . $self->coda_dir . "/CODA\")" : "";

	assert($self->model_file, "file name for model must be specified");
	assert($self->data_file,
		   "file name for data must be set before script()");
	assert($self->init_file,
		   "file name for initial values must be set before script()");

	push @sc, "# jags script\n";
	push @sc, "model in \"" . $self->model_file . "\"\n";
	push @sc, "data in \"" . $self->data_file->{name} . "\"\n";
	push @sc, "compile, nchains(" . $self->nchains . ")\n";
	push @sc, "parameters in \"" . $self->init_file->{name} . "\"\n";
	push @sc, "initialize\n";
	push @sc, "$burnin " . $self->burnin . "\n" if $self->burnin;
	for (sort keys %{$self->_monitor}) {
		push @sc, "monitor $_\n";
	}
	push @sc, "update " . $self->update . "\n" if $self->update;
	push @sc, "coda " . $self->coda_arg . "$stem\n" if $self->coda_arg;
	push @sc, "parameters to \"" . $self->param_file . "\"\n"
	  if $self->param_file;
	push @sc, "samplers to \"" . $self->sampler_file . "\"\n"
	  if $self->sampler_file;
	push @sc, "# end of script\n";

	return \@sc;
}

# _build_coda: default initialize of 'coda'
sub _build_coda {
	my $self = shift;
	my @c = ();

	for my $c (1 .. $self->nchains) {
#		print "build coda: $c\n";
		$c[$c] = Makeshift::Jags::coda->new(chain => $c,
											coda_dir => $self->tmpdir);
	}

	return \@c;
}

#__PACKAGE__->meta->make_immutable();

1;
# end
