# perl interface for jags (just another Gibbs sampler)
# '17/07/05 by KGK

# analyze coda file
package Makeshift::Jags::coda;

use Carp::Assert;
use Mouse;
use Mouse::Util;
use File::Temp qw/ tempfile /;

my @stats = ('mean', 'sd', 'naive_se', 'time_series_se');

has 'R_command' => (is => 'rw', default => "Rscript");
has 'chain' => (is => 'rw', default => 1);
has 'coda_dir' => (is => 'ro', required => 1);
has ['R_file', 'log_file', 'conv_log']
  => (is => 'ro', lazy => 1, builder => '_handle_name');
has 'diag_method' => (is => 'rw', default => ''); # NULL or geweke
has 'converge' => (is => 'rw', lazy_build => 1);
has 'script' => (is => 'rw', lazy_build => 1);
has 'stem' => (is => 'rw', default => "CODA");
has 'names' => (is => 'ro', default => sub{[]});
has \@stats => (is => 'rw', default => sub{{}});

__PACKAGE__->meta->make_immutable();

# summary: show result in brief
sub summary {
	my $self = shift;
	my @label = @stats;
	my @result;

	$self->read_log if @{$self->names} == 0;

	push @label, $self->diag_method if $self->diag_method;
	print join(", ", ("# var", @label)), "\n";
	for my $var (@{$self->names}) {
		@result = ($var);
		for (@label) {
			my $val = ($_ eq $self->diag_method) ?
			  $self->converge->value->{$var} : $self->$_->{$var};
			push @result, $val;
		}
		print join(", ", @result), "\n";
	}

	if ($self->diag_method) {
		my $conv = $self->converge;
		if ($conv->check) {
			print "All parameters have converged\n";
		} else {
			for (@{$conv->names}) {
				print "Fail convergence for $_\n"
				  unless $conv->ok($conv->value->{$_});
			}
		}
	}
}

# read_log: read logfile and store result of monitored variables
sub read_log {
	my $self = shift;
	my $log = $self->log_file->{name};
	my $fh = $self->log_file->{handle};
	my %names = ();

	$self->run unless -s $log;

	assert(-f $log, "$log is not a file");
	assert(-s $log, "$log is empty");
	assert(-r $log, "$log is not readable");

# variables
	while (<$fh>) {
		last if /^summary:/;
		push @{$self->names}, $1 if /Abstracting\s+(\S+)\s+/;
	}
	for (@{$self->names}) {
		$names{$_} = 1;
	}
# summary
	while (<$fh>) {
		last if /^2\./;
		chomp;
		my @line = split;
		my $var = shift @line;
		next unless $var and $names{$var};

		for (@stats) {
			$self->$_->{$var} = shift @line;
		}
	}
# convergence diagnostic
	$self->converge->read_log($self->conv_log) if $self->diag_method;

	return $self;
}

# script_to(Rfile): write R script file
# Rfile[optional]: R script file
sub script_to {
    my $self = shift;
	my $Rfile = shift;

	$self->R_file($self->handle_name($Rfile)) if $Rfile;
	assert($self->R_file, "file name for R-script must be specified");

	my $fh = $self->R_file->{handle};
	for (@{$self->script}) {
		print $fh $_;
	}

    return $self->R_file->{name};
}

# run: run Rscript
sub run {
    my $self = shift;

	system $self->R_command . " " . $self->script_to;

	die $self->R_command . " return $?" if $?;

    return $self;
}

# _handle_name: return hash reference for handle and name for tmpfile
sub _handle_name {
	my $self = shift;
	my ($fh, $fn) = tempfile(DIR => $self->coda_dir);

	return {handle => $fh, name => $fn};
}

# _build_script: default initialize of 'script'
# create standard R script as strings array
sub _build_script {
    my $self = shift;
	my @sc = ();
	my $base = $self->coda_dir . "/" . $self->stem;
	my $index = $base . "index.txt";
	my $chain = $base . "chain" . $self->chain . ".txt";
	my $log = $self->log_file->{name};

	assert(-f $index, "$index is not a file");
	assert(-r $index, "$index is not readable");
	assert(-f $chain, "$chain is not a file");
	assert(-r $chain, "$chain is not readable");

	push @sc, "# R script\n";
	push @sc, "library(coda)\n\n";
	push @sc, "sink(\"$log\")\n";
	push @sc, "codadata <- read.coda(\"$chain\", \"$index\")\n";
	push @sc, "cat(\"summary:\")\n";
	push @sc, "summary(codadata)\n";
	if ($self->diag_method) {
		my $log = $self->conv_log->{name};
		my $conv = $self->converge;
		my $method = $conv->method;
		push @sc, "sink(\"$log\")\n";
		push @sc, $conv->R_line("codadata") . "\n";
	}
	push @sc, "\n";
	push @sc, "# end of script\n";

	return \@sc;
}

# _build_converge: default initialize of 'converge'
# for convergence diagnostic
sub _build_converge {
    my $self = shift;
	my $method = $self->diag_method;

	assert($method, "diag_method must be specified");

	my $class = "Makeshift::Jags::coda::$method";
	Mouse::Util::load_class($class);
	my $conv = $class->new;

	return $conv;
}

# __PACKAGE__->meta->make_immutable();

1;
# end
